#!/usr/bin/python3
#  coding: utf-8
#
#       setTempPress_getTemp.py
#
#       OpenFAOMの熱流速heatFluxをFrontISTRのcntファイルにセットし、
#       温度の点群ファイルを作成する。
#       温度の点群ファイルからOFのpointにmappingした結果fileを
#       作成する。（OF側がこのfileを読み込む）
#
#   23/04/04    新規作成
#      05/04    温度のmappingを修正
#

import sys
import os
import time
import glob
import mmap
import mappingFromPointCloud as mapping
import pyTreeFoam_ptc as pyTreeFoam
import createPointCloudFromOpenFoam as createPtc
import pyCoupling

waitCount = 5000
waitTime = 0.01

commonSize = 10000


#
#  getMappedFacePointLocs
#---------------------
def getMappedFacePointLocs():
    """ mappingの為の辞書を作成"""
    #mapPointDictを取得
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapPointDict = {}
    for line in lines:
        words = line.split()
        if len(words) > 4:
            nodeNo = int(words[0])
            if not (nodeNo in mapPointDict.keys()):
                pointLoc = list(map(float, words[1:4]))
                mapPointDict[nodeNo] = pointLoc
    #mapFaceCenterDictを作成
    name = patchName + "_mapFaceCenterNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapFaceCenterDict = {}
    for line in lines:
        words = line.split()
        if len(words) >= 7:
            faceNo = int(words[0])
            faceLoc = list(map(float, words[1:4]))
            nodeNos = list(map(int, words[4:]))
            mapFaceCenterDict[faceNo] = [faceLoc, nodeNos]
    return mapFaceCenterDict, mapPointDict

#
# createPtcFileFromOF
#----------------------
def createPtcFileFromOF(valueFile, dataDict):
    """ 点群file作成"""
    fluidCaseDir = dataDict["fluidCaseDir"]
    patchName = dataDict["patchName"]
    procNo = dataDict["procNo"]
    #faceCenterLoc読み込み
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceCenterFile = dataDir + "/" + patchName + "_faceCenterLocs_" + procNo
    f = open(faceCenterFile); lines = f.readlines(); f.close()
    #faceValueを読み込み
    waitUntilValueFile(valueFile)
    fp = open(valueFile); pLines = fp.readlines(); fp.close()
    newLines = []
    for i in range(len(lines)):
        newLine = lines[i][:-1] + " " + pLines[i]
        newLines.append(newLine)
    newLines.append("#end\n")
    #streamにセットし、全体制御側に送信
    conts = "".join(newLines)
    return conts

#
#  waitUntilValueFile
def waitUntilValueFile(valueFile):
    """ valueFileが完成するまで待つ"""
    count = waitCount
    sleepTime = waitTime
    while True:
        if os.path.exists(valueFile) == True:
            time.sleep(sleepTime)
            break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  saveMappingTemp
#-------------------
def saveMappingTemp(mappingDict):
    """ mappingDispの結果を保存。このファイルをOF側が読み込む"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceNos = list(mappingDict.keys())
    faceNos.sort()
    lines = []
    for faceNo in faceNos:
        temp = mappingDict[faceNo]
        line = str(temp) + "\n"
        lines.append(line)
    #保存のfilename
    name = patchName + "_temp_" + beforeTime + "_" + procNo
    tempFile = saveDir + "/" + name
    f = open(tempFile, "w"); f.writelines(lines); f.close()

#
#  getLocTemperature
#--------------------
def getLocTemperature(lines):
    """ 温度の点群ファイルを取得"""
    locTemp = []
    for line in lines:
        if line[0] != "#":
            words = line.split()
            loc = list(map(float, words[:3]))
            temp = float(words[3])
            locTemp.append([loc, temp])
    return locTemp

#
#  isMappingTriFile
def isMappingTriFile():
    """ mapping用の3点を取得したfileが存在するかチェック"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    return os.path.exists(fileName)

#
#  createMappingTriFile
def createMappingTriFile(mappingTriDict):
    """ mapping用の3点の座標をfileに保存する"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    lines = []
    for key in mappingTriDict.keys():
        faceNo = key
        word = str(faceNo)
        triLocs = mappingTriDict[key]
        loc0, loc1, loc2 = triLocs
        word0 = " ".join(list(map(str, loc0)))
        word1 = " ".join(list(map(str, loc1)))
        word2 = " ".join(list(map(str, loc2)))
        line = " ".join([word, word0, word1, word2]) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

#
#  readMappingTriFile
def readMappingTriFile():
    """ 保存したmapping用の3点の座標を読み込む"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mappingTriDict = {}
    for line in lines:
        words = line.split()
        faceNo = int(words[0])
        key = faceNo
        loc0 = list(map(float, words[1:4]))
        loc1 = list(map(float, words[4:7]))
        loc2 = list(map(float, words[7:10]))
        mappingTriDict[key] = [loc0, loc1, loc2]
    return mappingTriDict


#-----------------------
#  setHeatFlux_getTemp
#-----------------------
def setHeatFlux_getTemp(dataDict):
    """ 熱流束をmappingしてfistrの結果を待ち、温度を取得する。"""
    solidCaseDir = dataDict["solidCaseDir"]
    fluidCaseDir = dataDict["fluidCaseDir"]
    patchName = dataDict["patchName"]
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    mesh = mapping.fistrMesh(solidCaseDir)
    #温度の点群ファイル作成
    ptcDir = fluidCaseDir + "/coupling_FrontISTR/data"
    htFile = ptcDir + "/" + patchName + "_faceHeatFlux_" + timeName + "_" + procNo
    htConts = createPtcFileFromOF(htFile, dataDict)
    #heatFluxを転送
    title = "heatFlux"
    streamToParent.put(title, htConts)
    #mappingするfaceCenterの座標を予め取得しておく（待ち時間を利用）
    mapFaceCenterDict, mapPointDict = getMappedFacePointLocs()
    #変位の点群dataをstreamから読み込み
    lines = streamToParent.getLines("temperature")
    locTemp = getLocTemperature(lines)

    # #mappingTriFile(mapping用の3点の座標を保存)があるか？
    # if isMappingTriFile() == False:    
    #     #存在しない場合、mappingTriDictを取得
    #     mapStat, mappingTriDict = mesh.getMappingTriLocs(locTemp, mapPointFaceDict, mapPointDict)
    #     if mapStat != "OK":
    #         return "error in temperature mapping."
    #     #mappingTriFileを作成
    #     createMappingTriFile(mappingTriDict)
    # else:
    #     #fileを読み込む
    #     mappingTriDict = readMappingTriFile()

    mapStat, mappingTriDict = mesh.getMappingTriLocs(locTemp, mapFaceCenterDict, mapPointDict)

    #温度をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locTemp, mapFaceCenterDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingTemp(mappingDict)
    return ""

#
#  beginning
#------------
def beginning(dataDict):
    """ beforeTime, currTimeをファイルに保存"""
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    #processor0のみ
    if procNo == "0":
        #時間を書き込み
        cont = pyCoupling.createTimeCont(beforeTime, timeName)
        # lines =  ["beforeTime " + beforeTime + "\n"]
        # lines += ["currTime   " + timeName + "\n"]
        # lines += ["coupling   yesno\n"]
        # lines += ["#end\n"]
        # cont = "".join(lines)
        mmTimes.seek(0)
        mmTimes.write(cont.encode())
        mmTimes.flush()
    #couplingFileに反映されるまで待つ
    waitUntilCouplingWrote(dataDict)
    return

#
#  waitUntilCouplingWrote
def waitUntilCouplingWrote(dataDict):
    """ couplingStepにbeforeTimeが反映されるまで待つ"""
    count = waitCount
    sleepTime = waitTime
    beforeTime = dataDict["beforeTime"]
    while True:
        mmCoupling.seek(0)
        cont = mmCoupling.read().decode()
        lines = cont.split("\n")
        if len(lines) >= 3:
            bTime = lines[0].split()[1]
            if float(bTime) == float(beforeTime):
                #beforeTimeが反映された場合戻る
                break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  isCouplingTime
#-----------------
def isCouplingTime(dataDict):
    """ couplingTime（連成させる時間）かどうか確認
    （timesの内容はmmCouplingStepに反映されている。）"""
    startTime = dataDict["startTime"]
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    nSteps = dataDict["nSteps"]
    writeInterval = dataDict["writeInterval"]
    coupling = False
    if int(nSteps) == 1:
        #NSTEPS=1の場合は、毎回連成
        coupling = True
        return coupling
    
    deltaT = str(float(timeName) - float(beforeTime))
    if pyCoupling.isWriteInterval(timeName, writeInterval, deltaT) == True:
        #writeIntervalでは連成する
        coupling = True
        return coupling
    #mmCouplingStep読込
    mmCoupling.seek(0)
    conts = mmCoupling.read().decode()
    lines = conts.split("\n")
    bTime = lines[0].split()[1]
    cTime = lines[1].split()[1]
    skipSteps = int(lines[2].split()[1])
    skipStepsTemp = int(lines[3].split()[1])
    if skipStepsTemp == 0:
        coupling = True
        return coupling
    if float(bTime) == 0:
        #初回は連成
        coupling = True
        return coupling
    if float(beforeTime) == float(startTime):
        coupling = True
        return coupling
    return coupling

#
#  createInitialTempFile
#------------------------
def createInitialTempFile(dataDict):
    """ 初期の温度結果fileを作成"""
    #patchの値を取得
    fluidCaseDir = dataDict["fluidCaseDir"]
    case = pyTreeFoam.case(fluidCaseDir)
    region = "."
    pMesh = case.getCurrMeshDir(time, region, "boundary")
    meshDir = fluidCaseDir + os.sep + pMesh.split(os.sep)[0]
    polyMesh = createPtc.ofMesh(fluidCaseDir, meshDir)
    patchName = dataDict["patchName"]
    startTime = "0"; field = "T"; patches = [patchName]
    patchDataDict = polyMesh.patchDataDictFromField(startTime, field, patches, region)
    faceNos = list(patchDataDict.keys())
    faceNos.sort()
    #fileを作成
    beforeTime = dataDict["beforeTime"]
    procNo = dataDict["procNo"]
    name = patchName + "_temp_" + beforeTime + "_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName, "w")
    for faceNo in faceNos:
        line = str(patchDataDict[faceNo]) + "\n"
        f.write(line)
    f.close()

#
#  getOlderTime
def getOlderTime(bTime):
    """ bTimeの1個古いtimeを取得する"""
    mmTimeList.seek(0)
    conts = mmTimeList.read().decode()
    strTimes = conts.split("\n")
    if bTime in strTimes:
        id = strTimes.index(bTime)
        oldTime = strTimes[id+1]
        try:
            float(oldTime)
        except:
            oldTime = "-1"
    else:
        oldTime = "-1"
    return oldTime

#
#  copyMapFaceTempFile
#----------------------
def copyMapFaceTempFile():
    """ mappingTempのファイルをrenameする。"""
    #計算結果があるか？
    newName = patchName + "_temp_" + beforeTime + "_" + procNo
    newFileName = fluidCaseDir + "/coupling_FrontISTR/data/" + newName
    if os.path.exists(newFileName) == True:
        return
    #beforeTimeの直前のtimeを取得
    oldTime = getOlderTime(beforeTime)
    oldName = patchName + "_temp_" + oldTime + "_" + procNo
    oldFileName = fluidCaseDir + "/coupling_FrontISTR/data/" + oldName
    #oldFileNameが存在するか？
    if os.path.exists(oldFileName) == False:
        #存在しない場合、最新のfileを取得
        name = patchName + "_temp_*_" + procNo
        pattern = fluidCaseDir + "/coupling_FrontISTR/data/" + name
        files = glob.glob(pattern)
        fileTimes = []
        for fileName in files:
            ft = fileName.split("_")[-2]
            fileTimes.append([float(ft), ft, fileName])
        fileTimes.sort()
        oldName = patchName + "_temp_" + fileTimes[-1][1] + "_" + procNo
        oldFileName = fluidCaseDir + "/coupling_FrontISTR/data/" + oldName
    os.rename(oldFileName, newFileName)

#
#  createMmapFile
#-----------------
def createMmapFile(fileName):
    """ mmapFileを作成する"""
    f = open(fileName, "r+b"); mmFile = mmap.mmap(f.fileno(), 0); f.close()
    return mmFile

#
#  readMmapLines
def readMmapLines(fileName):
    """ fileObjectの共有メモリからlinesを読み込み返す"""
    f = open(fileName, "r+b")
    mm = mmap.mmap(f.fileno(), 0)
    f.close()
    lines = []
    while True:
        line = mm.readline().decode()
        if line == "":
            break
        lines.append(line)
    mm.close()
    return lines



if __name__ == "__main__":
    import gettext
    gettext.install("app") # replace with the appropriate catalog name
    _ = gettext.gettext

    timeName = sys.argv[1]      #currTime
    beforeTime = sys.argv[2]    #前回のtime
    patchName = sys.argv[3]     #patchName
    procNo = sys.argv[4]        #procNo
    nProcs = sys.argv[5]        #nProcs
    fluidCaseDir = os.getcwd()
    couplingData = pyCoupling.couplingData(fluidCaseDir)
    couplingDataDict = couplingData.read()
    couplingDataDict["fluidCaseDir"] = fluidCaseDir
    couplingDataDict["timeName"] = timeName
    couplingDataDict["beforeTime"] = beforeTime
    couplingDataDict["patchName"] = patchName
    couplingDataDict["procNo"] = procNo
    couplingDataDict["nProcs"] = nProcs
    waitCount = int(couplingDataDict["maxWaitTime"])
    #共有メモリの設定
    commonSize = int(couplingDataDict["streamBuffer"])
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    mmCommFile = createMmapFile(dataDir + "/common_" + procNo)
    mmTimes = createMmapFile(dataDir + "/times")
    mmCoupling = createMmapFile(dataDir + "/couplingStep")
    mmTimeList = createMmapFile(dataDir + "/timeList")
    #streamを定義
    streamToParent = pyCoupling.stream(mmCommFile, commonSize,
                                       waitCount, waitTime )
    beginning(couplingDataDict)
    #連成させる時間？
    if isCouplingTime(couplingDataDict) == False:
        #連成させない時間の場合、前回の結果をコピーして終了。
        copyMapFaceTempFile()
        exit()
    #計算開始時？（初期化の有無）
    elif (float(beforeTime) == float(couplingDataDict["startTime"]) 
        and float(couplingDataDict["startTime"]) == 0.0):
        #初回起動時の設定
        pass
    #途中から再開した場合（latestTime）
    elif (float(beforeTime) == float(couplingDataDict["startTime"])
        and int(couplingDataDict["nSteps"]) > 1):
        #copyMapFaceTempFile()
        #exit()
        pass
    #連続計算
    solidCaseDir = os.path.abspath(couplingDataDict["solidCaseDir"])
    stat = setHeatFlux_getTemp(couplingDataDict)
    if stat != "":
        raise NameError(stat)

    